# -*- coding:utf-8 -*-
#!/usr/bin/python3
from PyQt6.QtWidgets import QPushButton
from PyQt6 import QtWidgets, QtGui, QtCore
from .dancingdots import OpenGLDancingDots
import time

class ColourButton(QPushButton):
    """
    Helper class. A button with a connected QColor dialog, that has a signal .valueChanged and a .value() function
    that returns three ints (the rgb value of the colour)
    """

    _value = (0, 0, 0)     # the current colour, a tuple of three ints (rgb 0/255)
    valueChanged = QtCore.pyqtSignal((int, int, int), name='valueChanged')   # signal when user selects colour

    def value(self):
        """
        :return: The current colour
        """
        return self._value

    def setValue(self, colour):
        """ Sets the current colour and changes the button background
        :param colour: The colour to set, a tuple of three ints
        """
        self._value = colour
        self.setStyleSheet('background-color: #{:02x}{:02x}{:02x}'.format(*self._value))

    def __init__(self):
        """ Connecting the the base class' signal
        """
        super().__init__()
        self.clicked.connect(self.select_colour)

    def select_colour(self):
        """ Callback when clicked; calls the colour dialog and stores the selected value
        :return: void
        """
        self.setValue(QtWidgets.QColorDialog.getColor(initial=QtGui.QColor(*self._value)).getRgb()[:3])
        self.valueChanged.emit(self._value[0], self._value[1], self._value[2])    # seperately, tuples aren't supported



class DancingDotsLayout(QtWidgets.QGridLayout):

    _showing_ddots = False
    _ddots_wdg = None       # opengl widget with the dots
    _settings_wdg = None     # widget with all settings
    _no_settings_wdg = None    # empty settins with a "show" button
    _settings = {}     # dictionary with settings widgets: 'string' → widget

    def __init__(self, parent, get_data):
        super().__init__(parent)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum)
        parent.setSizePolicy(sizePolicy)

        # add setting layout
        self._settings_wdg = QtWidgets.QWidget()
        self._no_settings_wdg = QtWidgets.QWidget()
        settingslayout = QtWidgets.QGridLayout(self._settings_wdg)
        no_settingslayout =  QtWidgets.QGridLayout(self._no_settings_wdg)

        # show/hide buttons
        show_settings_btn = QtWidgets.QPushButton("☰")
        show_settings_btn.setStyleSheet("padding: 0.3em; font-size: 12pt;")
        show_settings_btn.clicked.connect(self.show_settings)
        hide_settings_btn = QtWidgets.QPushButton("Hide Settings")
        hide_settings_btn.clicked.connect(self.hide_settings)
        settingslayout.addWidget(hide_settings_btn, 0, 0, 1, 4)
        no_settingslayout.addWidget(show_settings_btn, 0, 0, 1, 1)
        settingslayout.addItem(QtWidgets.QSpacerItem(10, 20,
                                                       QtWidgets.QSizePolicy.Policy.Minimum,
                                                       QtWidgets.QSizePolicy.Policy.Expanding),
                                 1, 0, 1, 4)
        no_settingslayout.addItem(QtWidgets.QSpacerItem(10, 20,
                                                           QtWidgets.QSizePolicy.Policy.Minimum,
                                                           QtWidgets.QSizePolicy.Policy.Expanding),
                                    1, 0, 1, 1)

        # controls for time smoothing and power
        # spacer
        settingslayout.addItem(QtWidgets.QSpacerItem(10, 20,
                                                     QtWidgets.QSizePolicy.Policy.Minimum,
                                                     QtWidgets.QSizePolicy.Policy.Expanding),
                               2, 0, 1, 4)

        # label
        settingslayout.addWidget(QtWidgets.QLabel("General"), 3, 0, 1, 4)

        # circular frequency
        spin_timesmoothing = QtWidgets.QDoubleSpinBox()
        spin_timesmoothing.setRange(0.0, 1.0)
        spin_timesmoothing.setSingleStep(0.02)
        spin_timesmoothing.valueChanged.connect(self.update_ddots_display)
        settingslayout.addWidget(spin_timesmoothing, 4, 0, 1, 1)
        settingslayout.addWidget(QtWidgets.QLabel("Time smoothing"), 4, 1, 1, 1)
        self._settings['timesmoothing'] = spin_timesmoothing

        # rotation
        spin_power = QtWidgets.QDoubleSpinBox()
        spin_power.setRange(0.0, 5.0)
        spin_power.setSingleStep(0.5)
        spin_power.valueChanged.connect(self.update_ddots_display)
        settingslayout.addWidget(spin_power, 4, 2, 1, 1)
        settingslayout.addWidget(QtWidgets.QLabel("Power/Softmax"), 4, 3, 1, 1)
        self._settings['power'] = spin_power

        # controls for repr. frequencies
        headings = ['Delta', 'Theta', 'Alpha', 'Beta', 'Gamma']
        for n in range(0, 5):

            # spacer
            settingslayout.addItem(QtWidgets.QSpacerItem(10, 20,
                                                         QtWidgets.QSizePolicy.Policy.Minimum,
                                                         QtWidgets.QSizePolicy.Policy.Expanding),
                                   4*n+5, 0, 1, 4)

            # label
            settingslayout.addWidget(QtWidgets.QLabel(headings[n]), 4*n+6, 0, 1, 4)

            # circular frequency
            spin_freq = QtWidgets.QSpinBox()
            spin_freq.setRange(0, 50)
            spin_freq.valueChanged.connect(self.update_ddots_display)
            settingslayout.addWidget(spin_freq, 4*n+7, 0, 1, 1)
            settingslayout.addWidget(QtWidgets.QLabel("Circular frequency"), 4*n+7, 1, 1, 1)
            self._settings['freq{}'.format(n)] = spin_freq

            # rotation
            spin_rotation = QtWidgets.QDoubleSpinBox()
            spin_rotation.setRange(-2.0, 2.0)
            spin_rotation.setSingleStep(0.1)
            spin_rotation.valueChanged.connect(self.update_ddots_display)
            settingslayout.addWidget(spin_rotation, 4*n+7, 2, 1, 1)
            settingslayout.addWidget(QtWidgets.QLabel("Rotation"), 4*n+7, 3, 1, 1)
            self._settings['rotation{}'.format(n)] = spin_rotation

            # inside colour
            cbutton_incolour = ColourButton()
            cbutton_incolour.valueChanged.connect(self.update_ddots_display)
            settingslayout.addWidget(cbutton_incolour, 4*n+8, 0, 1, 1)
            settingslayout.addWidget(QtWidgets.QLabel("Inside colour"), 4*n+8, 1, 1, 1)
            self._settings['incolour{}'.format(n)] = cbutton_incolour

            # outside colour
            cbutton_outcolour = ColourButton()
            cbutton_outcolour.valueChanged.connect(self.update_ddots_display)
            settingslayout.addWidget(cbutton_outcolour, 4*n+8, 2, 1, 1)
            settingslayout.addWidget(QtWidgets.QLabel("Outside colour"), 4*n+8, 3, 1, 1)
            self._settings['outcolour{}'.format(n)] = cbutton_outcolour

        # add default settings
        settings = {
            'timesmoothing': 0.95, 'power': 2.0,
            'freq0':  2, 'rotation0':  0.7, 'incolour0': (255,   0,   0), 'outcolour0': ( 0, 64, 64),
            'freq1':  3, 'rotation1': -1.0, 'incolour1': (255, 128,   0), 'outcolour1': ( 0, 32, 64),
            'freq2':  4, 'rotation2':  0.6, 'incolour2': (255, 255,   0), 'outcolour2': ( 0,  0, 64),
            'freq3':  8, 'rotation3': -0.8, 'incolour3': (  0, 255,   0), 'outcolour3': (64,  0, 64),
            'freq4': 13, 'rotation4':  0.5, 'incolour4': (  0, 255, 128), 'outcolour4': (64,  0, 32),
        }
        self.set_settings(settings)

        # add widgets to main layout
        self.addWidget(self._no_settings_wdg, 0, 0, 1, 1)
        self._ddots_wdg = OpenGLDancingDots(get_data, self.toggle_fullscreen_dancing_dots, settings)
        self.addWidget(self._ddots_wdg, 0, 1, 1, 1)
        self.setColumnStretch(0, 0)
        self.setColumnStretch(1, 1)

        # start display thread
        time.sleep(1)
        self._showing_ddots = True
        self._ddots_wdg.start()

    def show_settings(self):
        """ Shows the settings pane in the layout
        :return: void
        """
        print("showing settings")
        self.setColumnStretch(0, 1)
        self.setColumnStretch(1, 5)
        self._no_settings_wdg.hide()
        self.removeWidget(self._no_settings_wdg)
        self._settings_wdg.show()
        self.addWidget(self._settings_wdg, 0, 0, 1, 1)

    def hide_settings(self):
        """ Hides the settings pane in the layout
        :return: void
        """
        print("hiding settings")
        self.setColumnStretch(0, 0)
        self.setColumnStretch(1, 1)
        self._settings_wdg.hide()
        self.removeWidget(self._settings_wdg)
        self._no_settings_wdg.show()
        self.addWidget(self._no_settings_wdg, 0, 0, 1, 1)

    def get_settings(self):
        """
        :return: A dictionary with setting names and the values from the GUI
        """
        settings = {}
        for name, widget in self._settings.items():
            settings[name] = widget.value()
        return settings

    def set_settings(self, settings):
        """
        :param settings: A dictionary with values to set
        """
        for name, value in settings.items():
            self._settings[name].setValue(value)

    def update_ddots_display(self):
        """ Set parameters in ddots display
        :return: void
        """
        if self._showing_ddots:
            self._ddots_wdg.set_parameters(self.get_settings())

    def toggle_fullscreen_dancing_dots(self, fullscreen):
        if not fullscreen:
            self.addWidget(self._ddots_wdg, 0, 1, 1, 1)
        if fullscreen:
            self.removeWidget(self._ddots_wdg)
            self._ddots_wdg.setParent(None)
            self._ddots_wdg.showFullScreen()




# -*- coding:utf-8 -*-
#!/usr/bin/python3
import ctypes

from OpenGL import GL as gl
from PyQt6 import QtCore, QtOpenGLWidgets, QtOpenGL
from PyQt6.QtGui import QSurfaceFormat
import numpy as np

class OpenGLDancingDots(QtOpenGLWidgets.QOpenGLWidget):

    _get_data = False
    _x_numbers = False
    _y_numbers = False
    _r_numbers = False
    _phi_numbers = False
    _vertices = False
    _red = False
    _green = False
    _blue = False
    _timer = False
    _p = 0.0
    _q = 0.0
    _r = []
    _k = []
    _shader_program_id = 0
    _vertex_buffer = None  # VBO
    _vertex_array = None  # VAO
    _vertices = False
    _counter = 0.0
    _M = 0
    _N = 0
    _softmax = 3.0
    _viewport = [0.0, 0.0, 0.0, 0.0]
    _update_viewport = False
    _fullscreen = False    # current display
    _toggle_fullscreen = None     # callback for onclick function

    # inside:  red / yellow / blue / orange / green
    # outside: orange / green / red / yellow / blue
    red = np.array([0.8, 0.0, 0.0])
    orange = np.array([1.0, 0.5, 0.0])
    yellow = np.array([0.6, 0.6, 0.0])
    green = np.array([0.0, 0.6, 0.0])
    turquoise = np.array([0.0, 0.6, 0.6])
    blue = np.array([0.0, 0.0, 1.0])
    purple = np.array([0.8, 0.0, 0.4])
    black = np.array([0.0, 0.0, 0.0])
    dark = 0.3
    _data_colours = [[red, dark*green],
                     [yellow, dark*blue],
                     [orange, dark*turquoise],
                     [green, dark*red],
                     [blue, dark*yellow]]
    _rotations = [0, 0, 0, 0, 0]

    def __init__(self, get_data, toggle_fullscreen, settings):
        super().__init__()

        self._get_data = get_data
        self._toggle_fullscreen = toggle_fullscreen

        self._M = 30  # number of circles
        self._N = 200  # points per circle

        # initialise data structures
        self._r_numbers = np.arange(0, self._M * 2.0 * np.pi, 2.0 * np.pi / self._N, dtype=np.float32)
        self._r_numbers = np.mod(self._r_numbers, 2.0 * np.pi)
        self._phi_numbers = np.sin(self._r_numbers, dtype=np.float32)
        self._x_numbers = np.zeros(self._r_numbers.shape, dtype=np.float32)
        self._y_numbers = np.zeros(self._r_numbers.shape, dtype=np.float32)
        self._red = np.zeros(self._r_numbers.shape, dtype=np.float32)
        self._green = np.zeros(self._r_numbers.shape, dtype=np.float32)
        self._blue = np.zeros(self._r_numbers.shape, dtype=np.float32)

        self.set_parameters(settings)


    def set_parameters(self, settings):

        if self._timer:
            self._timer.stop()

        # k direction
        # p in/out movement
        self._p = 0.0
        self._q = 1.0
        self._k = np.array([0.0, 0.0, 0.0, 0.0, 0.0])

        self._softmax = settings['power']
        for n in range(0, 5):

            # colours
            self._data_colours[n][0] = np.array(settings['incolour{}'.format(n)])/255.0
            self._data_colours[n][1] = np.array(settings['outcolour{}'.format(n)])/255.0

            # rotations
            self._rotations[n] = np.array(settings['rotation{}'.format(n)])

        # Create sum of sine waves of different frequencies
        dt = 0.005
        t0 = np.arange(0, 5 * np.pi, dt)
        sp = 1  #settings['shape']
        f = [(abs(np.sin(0.5 * settings['freq0'] * t0))**sp),
             (abs(np.sin(0.5 * settings['freq1'] * t0))**sp),
             (abs(np.sin(0.5 * settings['freq2'] * t0))**sp),
             (abs(np.sin(0.5 * settings['freq3'] * t0))**sp),
             (abs(np.sin(0.5 * settings['freq4'] * t0))**sp),]

        # Create 10 circles of different lengths
        freq_start = 1.0
        freq_step = 0.005
        self._r = [[]] * self._M  # the interpolated circle data, zeros for now, consisting of M*N circles s[M][1..5], all the same length
        s = [[]] * self._M        # the raw circle data, zeros for now, consisting of M*N circles s[M][1..5]
        t = [[]] * self._M        # the base to plot against, will go from 0 to 2pi, t[M]
        for m in range(0, self._M):
            s[m] = [[]] * 5
            freq = freq_start + m * freq_step  # our circle frequency
            for n in range(0, 5):
                s[m][n] = np.zeros(int((2 / freq) * np.pi / dt))
            t[m] = [m for m in np.arange(0, 2 * np.pi, 2 * np.pi / self._N)]

        # Average over all circles - pre-calculate parts of the sum of sines
        for m in range(0, self._M):
            for n in range(0, 5):
                # only plot if you can fill the entire circle
                for i in range(0, (len(f[n]) // len(s[m][n])) * len(s[m][n])):
                    value = f[n][i] - 0.5
                    index = i % len(s[m][n])
                    s[m][n][index] += value  # add to front
                    s[m][n][-(index + 1)] += value  # add to back, so beginning and end match and the circle stays closed

        # downsample so all circles have the same length (N samples)
        for m in range(0, self._M):
            self._r[m] = [[]] * 5
            for n in range(0, 5):
                self._r[m][n] = np.zeros(self._N)
                for i in range(0, self._N - 1):
                    index = int(float(i) * (len(s[m][n]) - 1) / (self._N - 1.0))
                    self._r[m][n][i] = s[m][n][index]
                self._r[m][n][self._N - 1] = s[m][n][0]  # close the circle

        # restart timer again
        if self._timer:
            self._timer.start(30)

    def initializeGL(self):

        print(f"GL_VENDOR: {gl.glGetString(gl.GL_VENDOR)}")
        print(f"GL_RENDERER: {gl.glGetString(gl.GL_RENDERER)}")
        print(f"GL_VERSION: {gl.glGetString(gl.GL_VERSION)}")
        print(f"GL_SHADING_LANGUAGE_VERSION: {gl.glGetString(gl.GL_SHADING_LANGUAGE_VERSION)}")

        # the vertex shader
        vertex_shader_id = gl.glCreateShader(gl.GL_VERTEX_SHADER)
        shader_code = (" #version 330 core\n"
                       " layout (location = 0) in vec2 xyCoords; "
                       " layout (location = 1) in vec3 vxColour; "
                       " out vec4 vertColour; "
                       " void main() { "
                       "     gl_Position = vec4(xyCoords, 0.0, 1.0); "
                       "     vertColour = vec4(vxColour, 1.0); "
                       " } ")
        gl.glShaderSource(vertex_shader_id, shader_code)
        gl.glCompileShader(vertex_shader_id)
        if gl.glGetShaderiv(vertex_shader_id, gl.GL_COMPILE_STATUS) == gl.GL_FALSE:
            print("Error creating dancing dots vertex shader.")

        # the fragment shader
        fragment_shader_id = gl.glCreateShader(gl.GL_FRAGMENT_SHADER)
        shader_code = (" #version 330 core\n"
                       " in vec4 vertColour; "
                       " out vec4 fragColour; "
                       " void main() { "
                       "     vec2 coords = gl_PointCoord * 2.0 - 1.0; "
                       "     float dist = length(coords); "
                       "     if (dist > 1.0) "
                       "         discard; "
                       "     else "
                       "         fragColour = vertColour; "
                       " } ")
        gl.glShaderSource(fragment_shader_id, shader_code)
        gl.glCompileShader(fragment_shader_id)
        if gl.glGetShaderiv(fragment_shader_id, gl.GL_COMPILE_STATUS) == gl.GL_FALSE:
            print("Error creating dancing dots fragment shader.")

        # the shader program, linking both shaders
        self._shader_program_id = gl.glCreateProgram()
        gl.glAttachShader(self._shader_program_id, vertex_shader_id)
        gl.glAttachShader(self._shader_program_id, fragment_shader_id)
        gl.glLinkProgram(self._shader_program_id)
        if gl.glGetProgramiv(self._shader_program_id, gl.GL_LINK_STATUS) == gl.GL_FALSE:
            print("Error linking dancing dots shaders.")
        if not gl.glIsProgram(self._shader_program_id):
            print(f"Error: Shader program {self._shader_program_id} is not valid!")

        # declare the buffer to be a vertex array
        self._vertices = np.column_stack((self._x_numbers, self._y_numbers, self._red, self._green, self._blue)).ravel()

        # create the VAO
        self._vertex_array = QtOpenGL.QOpenGLVertexArrayObject()
        self._vertex_array.create()
        self._vertex_array.bind()

        self._vertex_buffer = QtOpenGL.QOpenGLBuffer()
        self._vertex_buffer.create()
        self._vertex_buffer.bind()
        gl.glBufferData(gl.GL_ARRAY_BUFFER, self._vertices.nbytes, self._vertices, gl.GL_DYNAMIC_DRAW)
        size = self._vertices.itemsize
        gl.glVertexAttribPointer(0, 2, gl.GL_FLOAT, gl.GL_FALSE, 5*size, ctypes.c_void_p(0))
        gl.glEnableVertexAttribArray(0)
        gl.glVertexAttribPointer(1, 3, gl.GL_FLOAT, gl.GL_FALSE, 5*size, ctypes.c_void_p(2 * size))
        gl.glEnableVertexAttribArray(1)
        gl.glUseProgram(self._shader_program_id)
        gl.glPointSize(6)

    def paintGL(self):

        # cX - amount of frequency ring fX for each frequency X (out of five)
        freqs = self._get_data()
        [c1, c2, c3, c4, c5] = (freqs**self._softmax) / (freqs**self._softmax).sum()
        c6 = 0.1 * (c1 - c2 + c3 - c4 + c5)  # c6 - amount of in/out movement
        print("Frequency bands: {:.2f}, {:.2f}, {:.2f}, {:.2f}, {:.2f}".format(c1, c2, c3, c4, c5), end='\r')

        # k[X] - turning speed and direction of each frequency ring
        self._k += 0.5*np.array([self._rotations[0]*c1, self._rotations[1]*c2, self._rotations[2]*c3,
                                 self._rotations[3]*c4, self._rotations[4]*c5])
        kn = np.floor(self._k).astype(int)  # left index into ring
        km = kn + 1  # right index into ring
        kp = km - self._k  # left amount
        kq = self._k - kn  # right amount

        # plot curves
        last_data = []
        for m in range(0, self._M):

            # offset = np.sqrt((M-m))
            offset = 1.3*(self._M - m)
            offset = np.sqrt(np.sqrt(offset * offset * offset)) / self._M # x^(3/4)

            # soft rolling of circles by interpolating between floor(k) and ceil(k)
            data = (c1 * (np.roll(self._r[m][0], kn[0]) * kp[0] + np.roll(self._r[m][0], km[0]) * kq[0])
                    + c2 * (np.roll(self._r[m][1], kn[1]) * kp[1] + np.roll(self._r[m][1], km[1]) * kq[1])
                    + c3 * (np.roll(self._r[m][2], kn[2]) * kp[2] + np.roll(self._r[m][2], km[2]) * kq[2])
                    + c4 * (np.roll(self._r[m][3], kn[3]) * kp[3] + np.roll(self._r[m][3], km[3]) * kq[3])
                    + c5 * (np.roll(self._r[m][4], kn[4]) * kp[4] + np.roll(self._r[m][4], km[4]) * kq[4])
                    + 1)
            data /= data.max()
            data *= offset
            data += offset
            try:
                if len(last_data):
                    fr = m*self._N         # from offset
                    to = (m + 1)*self._N   # to offset
                    self._phi_numbers[fr:to] = self._p * last_data + self._q * data
                    colour_in =   c1 * self._data_colours[0][0] \
                                + c2 * self._data_colours[1][0] \
                                + c3 * self._data_colours[2][0] \
                                + c4 * self._data_colours[3][0] \
                                + c5 * self._data_colours[4][0]
                    colour_out =  c1 * self._data_colours[0][1] \
                                + c2 * self._data_colours[1][1] \
                                + c3 * self._data_colours[2][1] \
                                + c4 * self._data_colours[3][1] \
                                + c5 * self._data_colours[4][1]
                    if m == 1:
                        self._red[fr:to] = colour_out[0] * self._q
                        self._green[fr:to] = colour_out[1] * self._q
                        self._blue[fr:to] = colour_out[2] * self._q
                    elif m == self._M - 1:
                        self._red[fr:to] = colour_in[0] * self._p
                        self._green[fr:to] = colour_in[1] * self._p
                        self._blue[fr:to] = colour_in[2] * self._p
                    else:
                        c_in = m / self._M
                        c_out = 1.0 - c_in
                        colour = c_in * colour_in + c_out * colour_out
                        self._red[fr:to] = colour[0]
                        self._green[fr:to] = colour[1]
                        self._blue[fr:to] = colour[2]
            except BaseException as e:
                print("Plot exception {} for curve n={}".format(e, m))
            last_data = data
        self._p = (self._p + c6) % 1
        self._q = 1.0 - self._p

        # convert to x/y/colour data and push to graphic card
        self._y_numbers = (self._phi_numbers * np.cos(self._r_numbers))
        self._x_numbers = self._phi_numbers * np.sin(self._r_numbers)
        self._vertices = np.column_stack((self._x_numbers, self._y_numbers, self._red, self._green, self._blue)).ravel()

        gl.glUseProgram(self._shader_program_id)
        self._vertex_array.bind()
        self._vertex_buffer.bind()
        gl.glBufferSubData(gl.GL_ARRAY_BUFFER, 0, self._vertices.nbytes, self._vertices)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT | gl.GL_STENCIL_BUFFER_BIT)
        gl.glClearColor(0.0, 0.0, 0.0, 1.0)
        if self._update_viewport:
            gl.glViewport(*self._viewport)

        # actual drawing
        gl.glDrawArrays(gl.GL_POINTS, 0, len(self._x_numbers))
        error = gl.glGetError()
        if error != gl.GL_NO_ERROR:
            print(f"glDrawArrays error: {error}")

    def resizeGL(self, width, height):
        size = min(width, height)
        x = (width - size) // 2
        y = (height - size) // 2
        gl.glUseProgram(self._shader_program_id)
        gl.glViewport(x, y, size, size)
        self._viewport = [x, y, size, size]
        self._update_viewport = True
        print(f"glViewport set to x={x}, y={y}, width={size}, height={size}")

    def start(self):
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self.update)
        self._timer.start(30)

    def mouseReleaseEvent(self, dummy):
        self._fullscreen = not self._fullscreen
        self._toggle_fullscreen(self._fullscreen)



# -*- coding:utf-8 -*-
#!/usr/bin/python3

import sys

from .mainwindow import *
from PyQt6 import QtCore, QtWidgets
from PyQt6.QtGui import QSurfaceFormat
import threading
import numpy as np
import time
from mne_lsl.stream import StreamLSL as Stream
from mne_lsl.lsl import resolve_streams
from matplotlib import use as mpl_use
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
import matplotlib.pyplot as plt
from .display.layouts import DancingDotsLayout

mpl_use("QtAgg")



class Mind:
    """
    Implements a complete data and calculation set of a single human.
    """

    # general
    _name = ""           # the person's name

    # data streaming related
    _streaming = False   # whether we're streaming currently
    _resolving = True    # whether we're looking for LSL streams
    _showing_eegpsd = False     # whether we're showing the eeg/psd tab
    _data_seconds = 2.0  # how much data do we have in the _eeg_data array
    _sampling_rate = 0   # sampling rate of eeg data
    _samples = 0         # seconds times sampling rate
    _fft_resolution = 0  # resolution (distance of one FFT bin to the next)
    _channels = 1        # number of channels in the data
    _history_length = 600.0   # length of history buffer in seconds
    _eeg_data = []       # pointer to the buffer that has just been filled, either data_a or data_b
    _eeg_lock = threading.Lock()    # lock for eeg data
    _eeg_times = []      # buffer containing eeg time stamps
    _fft_data = []       # buffer containing the fft
    _fft_lock = threading.Lock()    # lock for fft data
    _fft_freqs = []      # buffer containing fft frequencies
    _bnd_data = []       # frequency band data: band frequencies
    _bnd_lock = threading.Lock()    # lock for bnd data
    _hst_data = []       # frequency band data: ring buffer that is constantly rooled
    _hst_lock = threading.Lock()    # lock for hst data
    _eeg_stream = None   # the lsl eeg input stream inlet, if in eeg mode
    _clc_stream = None   # the lsl calculation output stream outlet, if in calculation mode

    # normalisation
    _bnd_smoothing = 0.95   #
    _bnd_max = []       # maximum values of bands, calculated from data d(t) by x(t+1) = max(0.99*x(t), d(t))
    _bnd_min = []       # minimum values of bands, calculated from data d(t) by x(t+1) = min(1.01*x(t), d(t))
    _bnd_mid = []       # the middle between max and min

    # main mind controls
    _combobox_streamname = False
    _lineedit_name = False
    _checkbox_connect_lsl = False
    _checkbox_display_eegpsd = False
    _checkbox_visualisation_ddots = False
    _parent_tabwidget = False
    _available_streams = []

    # info labels
    _gui_lock = threading.Lock()
    _bnd_label = False
    _bnd_info = ""
    _lsl_label = False
    _lsl_info = ""
    _eeg_label = False
    _eeg_info = ""

    # eeg display research controls → put into new class soon
    _eegpsd_layout = False
    _eegpsd_tab = False
    _eeg_axes = False
    _eeg_canvas = False
    _eeg_channel_height = 70e-6
    _fft_axes = False
    _fft_canvas = False
    _fft_channel_height = 50e-4
    _bnd_axes = False
    _bnd_canvas = False
    _bnd_height = 110e-4
    _hst_axes = False
    _hst_canvas = False
    _hst_height = 110e-4

    # dancing dots display
    _ddots_tab = False

    def __init__(self, combobox_streamname, lineedit_name, checkbox_connect,
                       checkbox_display_eegpsd, checkbox_visualisation_ddots,
                       lsl_info, eeg_info, bnd_info,
                       parent_tabwidget):
        self._combobox_streamname = combobox_streamname
        self._lineedit_name = lineedit_name
        self._checkbox_connect_lsl = checkbox_connect
        self._checkbox_display_eegpsd = checkbox_display_eegpsd
        self._checkbox_visualisation_ddots = checkbox_visualisation_ddots
        self._lsl_label = lsl_info
        self._eeg_label = eeg_info
        self._bnd_label = bnd_info
        self._parent_tabwidget = parent_tabwidget

        # Connect slot eeg stream
        self._checkbox_connect_lsl.clicked.connect(self._connect_eeg_stream)
        self._checkbox_display_eegpsd.clicked.connect(self._create_eegpsd_tab)
        self._checkbox_visualisation_ddots.clicked.connect(self._create_dancing_dots_tab)

        # start stream searching thread (extra, otherwise it blocks to GUI)
        thlsl = threading.Thread(target=self._find_sources)
        thlsl.start()

        # start gui update thread, filling stream selection box
        self._gui_timer = QtCore.QTimer()
        self._gui_timer.timeout.connect(self._update_gui)
        self._gui_timer.start(300)

    def __del__(self):
        self._connect_eeg_stream(False, restart_resolve=False)

    def _reset(self):
        """
        Resets the values after a stream disconnect, also removes the tab
        :return: void
        """

        # set streaming to false, making all threads stop
        self._streaming = False
        self._showing_eegpsd = False
        self._sampling_rate = 0  # sampling rate of eeg data
        self._samples = 0
        self._fft_resolution = 0
        self._channels = 1
        self._history_length = 600.0
        self._eeg_data = []
        try:
            self._eeg_lock.release()
        except:
            pass
        self._eeg_times = []
        self._fft_data = []
        try:
            self._fft_lock.release()
        except:
            pass
        self._fft_freqs = []
        self._bnd_data = []
        try:
            self._bnd_lock.release()
        except:
            pass
        self._hst_data = []
        try:
            self._hst_lock.release()
        except:
            pass
        self._eeg_stream = None
        self._clc_stream = None

        # set GUI elements
        try:
            self._checkbox_display_eegpsd.setEnabled(False)
            self._checkbox_display_eegpsd.setChecked(False)
            self._checkbox_visualisation_ddots.setEnabled(False)
            self._checkbox_visualisation_ddots.setChecked(False)
            self._lsl_label.setEnabled(False)
            self._checkbox_connect_lsl.setText("Click to connect")
            self._eeg_label.setEnabled(False)
            self._bnd_label.setEnabled(False)
            self._create_eegpsd_tab(False)
            self._create_dancing_dots_tab(False)
        except:
            pass

    def _connect_eeg_stream(self, connect, restart_resolve=True):
        """
        Connects a single EEG stream
        Add the stream 'name' to the array of streams and starts pulling data
        :return: void
        """

        # if we're connecting
        if connect and not self._streaming:

            # find what's selected
            stream_name = self._combobox_streamname.currentText()
            stream_type = stream_name[0:3]
            stream_name = stream_name[4:]
            self._name = self._lineedit_name.text() or "No Name"
            if stream_name:

                # open LSL stream
                if stream_type == 'LSL':

                    # first resolve an EEG stream on the lab network
                    print("Connecting to LSL stream... ")
                    streams = resolve_streams()

                    # create a new inlet to read from the stream
                    for s in streams:
                        s_name, s_id, s_channels, s_rate = stream_name.split(' | ')
                        if s.source_id == s_id:

                            # set gui info
                            self._channels = s.n_channels
                            self._sampling_rate = s.sfreq
                            self._samples = int(self._data_seconds * self._sampling_rate)
                            self._eeg_stream = Stream(self._data_seconds, name=s_name, stype=s.stype, source_id=s.source_id)
                            self._checkbox_connect_lsl.setText("Connected")

                            # connect to stream
                            self._eeg_stream.connect(acquisition_delay=0.1, processing_flags="all")
                            self._streaming = True

                            # start data reading thread
                            thstr = threading.Thread(target=self._read_lsl)
                            thstr.start()

                if stream_type == 'SML':

                    # set gui info
                    print("Starting data simulation / self-test...")
                    self._channels = 5
                    self._sampling_rate = 250
                    self._samples = int(self._data_seconds * self._sampling_rate)
                    self._checkbox_connect_lsl.setText("Connected")
                    # start simulation thread
                    self._streaming = True
                    thstr = threading.Thread(target=self._simulate_eeg)
                    thstr.start()

                # start analysis thread
                thanal = threading.Thread(target=self._analyse_psd)
                thanal.start()

                # enable checkbox
                self._checkbox_display_eegpsd.setEnabled(True)
                self._checkbox_visualisation_ddots.setEnabled(True)
                print("... Data source connected.")

        # if we're disconnecting
        elif not connect and self._streaming:

            print("Disconnecting from LSL stream... ")
            self._streaming = False   # stop the display threads before we disconnect the stream
            self._eeg_stream and self._eeg_stream.disconnect()
            self._reset()
            print("... LSL stream disconnected.")

            # start looking for streams again
            if restart_resolve:
                self._resolving = True
                thlsl = threading.Thread(target=self._find_sources)
                thlsl.start()
            else:
                self._resolving = False

    def _create_eegpsd_tab(self, create):

        if create:

            # note
            print("Creating EEG/PSD display tab.")

            # create widgets
            self._eegpsd_tab = QtWidgets.QWidget()
            self._eegpsd_layout = QtWidgets.QGridLayout(self._eegpsd_tab)
            sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Maximum, QtWidgets.QSizePolicy.Policy.Maximum)
            self._eegpsd_tab.setSizePolicy(sizePolicy)
            self._parent_tabwidget.addTab(self._eegpsd_tab, "")
            self._parent_tabwidget.setTabText(self._parent_tabwidget.indexOf(self._eegpsd_tab),
                                              self._name + " -- EEG / Spectrum")

            # channel names
            c_names = ((self._eeg_stream and self._eeg_stream.ch_names)
                       or ['C{}'.format(n+1) for n in range(0, self._channels)])

            # colours
            frame_c = (0.25, 0.25, 0.25)
            background = self._parent_tabwidget.parent().parent().palette().base().color()
            brightness = sum(background.getRgb()[:3]) / (3 * 255.0)
            outer_c = background.name()
            passepartout_c = self._parent_tabwidget.parent().parent().palette().button().color().name() #(0.25, 0.25, 0.25)
            if brightness < 0.5:
                label_c = (0.8, 0.8, 0.8)
                title_c = (0.9, 0.9, 0.9)
            else:
                label_c = (0.2, 0.2, 0.2)
                title_c = (0.1, 0.1, 0.1)

            # first eeg plot
            figure = plt.figure()
            self._eeg_canvas = FigureCanvasQTAgg(figure)
            self._eeg_axes = figure.add_subplot(111)
            self._eegpsd_layout.addWidget(self._eeg_canvas, 0, 0, 1, 1)
            self._eeg_axes.set_ylim(bottom=-0.2, top=self._channels + 1.2)
            plt.subplots_adjust(top=0.95, bottom=0.05, left=0.05, right=0.99)
            self._eeg_axes.set_xticks([])
            self._eeg_axes.set_yticks(ticks=np.arange(1, self._channels + 1), labels=c_names, color=label_c)
            self._eeg_axes.set_title('{} -- EEG over {:0.1f} Seconds'.format(self._name, self._data_seconds),
                                     color=title_c, fontsize=10, pad=5)
            self._eeg_axes.set_facecolor(outer_c)
            figure.set_facecolor(passepartout_c)
            plt.setp(self._eeg_axes.spines.values(), color=frame_c)

            # first psd plot
            figure = plt.figure()
            self._fft_canvas = FigureCanvasQTAgg(figure)
            self._fft_axes = figure.add_subplot(111)
            plt.subplots_adjust(top=0.95, bottom=0.05, left=0.1, right=0.99)
            self._eegpsd_layout.addWidget(self._fft_canvas, 0, 1, 1, 1)
            self._fft_axes.set_ylim(bottom=0.8, top=self._channels + 2.2)
            self._fft_axes.set_xscale('log')
            self._fft_axes.set_yticks(ticks=np.arange(1, self._channels + 1), labels=c_names, color=label_c)
            self._fft_axes.set_title('Current PSD', color=title_c, fontsize=10, pad=5)
            self._fft_axes.set_facecolor(outer_c)
            figure.set_facecolor(passepartout_c)
            plt.setp(self._fft_axes.spines.values(), color=frame_c)

            # bandpass history plot
            figure = plt.figure()
            self._hst_canvas = FigureCanvasQTAgg(figure)
            self._hst_axes = figure.add_subplot(111)
            plt.subplots_adjust(top=0.9, bottom=0.05, left=0.05, right=0.99)
            self._eegpsd_layout.addWidget(self._hst_canvas, 1, 0, 1, 1)
            self._hst_axes.set_ylim([-0.1, 5.1])
            self._hst_axes.set_xticks([])
            self._hst_axes.set_yticks([0, 1, 2, 3, 4], ['δ', 'θ', 'α', 'β', 'γ'], color=label_c)
            self._hst_axes.set_title('{} -- PSD History over {} minutes'.format(self._name, self._history_length/60.0),
                                     color=title_c, fontsize=10, pad=5)
            self._hst_axes.set_facecolor(outer_c)
            figure.set_facecolor(passepartout_c)
            plt.setp(self._hst_axes.spines.values(), color=frame_c)

            # bandpass bar graph
            figure = plt.figure()
            self._bnd_canvas = FigureCanvasQTAgg(figure)
            self._bnd_axes = figure.add_subplot(111)
            plt.subplots_adjust(top=0.9, bottom=0.1, left=0.1, right=0.99)
            self._eegpsd_layout.addWidget(self._bnd_canvas, 1, 1, 1, 1)
            self._bnd_axes.set_ylim([0.0, 1.1])
            self._bnd_axes.set_xticks([1, 2, 3, 4, 5], ['δ', 'θ', 'α', 'β', 'γ'], color=label_c)
            self._bnd_axes.set_yticks([])
            self._bnd_axes.set_title('Frequency Band Power'.format(self._data_seconds),
                                     color=title_c, fontsize=10, pad=5)
            self._bnd_axes.set_facecolor(outer_c)
            figure.set_facecolor(passepartout_c)
            plt.setp(self._bnd_axes.spines.values(), color=frame_c)

            self._eegpsd_layout.setColumnStretch(0, 3)
            self._eegpsd_layout.setColumnStretch(1, 1)
            self._eegpsd_layout.setRowStretch(0, 2)
            self._eegpsd_layout.setRowStretch(1, 1)

            # start display thread
            time.sleep(1)
            self._showing_eegpsd = True
            thdsp = threading.Thread(target=self._display_eeg_psd)
            thdsp.start()

        else:
            if self._eegpsd_tab:
                print("Removing EEG/PSD display tab.")
                self._parent_tabwidget.removeTab(self._parent_tabwidget.indexOf(self._eegpsd_tab))
            self._showing_eegpsd = False

    def _create_dancing_dots_tab(self, create):

        if create:

            # note
            print("Creating Dancing Dots display tab.")

            # create tab
            self._ddots_tab = QtWidgets.QWidget()

            # add dancing dot display layout to tab
            DancingDotsLayout(self._ddots_tab, self.get_data)
            self._parent_tabwidget.addTab(self._ddots_tab, "")
            self._parent_tabwidget.setTabText(self._parent_tabwidget.indexOf(self._ddots_tab),
                                              self._name + " -- Dancing Dots")

        else:
            if self._ddots_tab:
                print("Removing Dancing Dots display tab.")
                self._parent_tabwidget.removeTab(self._parent_tabwidget.indexOf(self._ddots_tab))

    def _display_eeg_psd(self):

        # starting thread
        print("Starting EEG/PSD display.")
        while not len(self._eeg_data)\
                or not len(self._fft_data)\
                or not len(self._bnd_data)\
                or not len(self._hst_data):
            time.sleep(0.2)

        # eeg + fft
        eeg_lines = self._eeg_axes.plot(self._eeg_data.T)  # the last channel in the simulator has the alpha intensity
        fft_lines = self._fft_axes.plot(self._fft_freqs, self._fft_data.T)
        bnd_bars = self._bnd_axes.bar([1, 2, 3, 4, 5], self._bnd_data)
        hst_lines = self._hst_axes.plot(self._hst_data.T)

        background = self._parent_tabwidget.parent().parent().palette().base().color()
        brightness = sum(background.getRgb()[:3]) / (3 * 255.0)

        # set rainbow colours for eeg and fft
        for c in range(0, len(eeg_lines)):
            a = c/self._channels
            if brightness < 0.5:
                colour = (0.3+0.7*(1 - a), 0.5+0.5*(1.0 - 2.0*abs(a - 0.5)), 0.3+0.7*a)
            else:
                colour = (0.7 - 0.7 * (1 - a), 0.5 - 0.5 * (1.0 - 2.0 * abs(a - 0.5)), 0.7 - 0.7 * a)
            eeg_lines[c].set_color(color=colour)
            eeg_lines[c].set_linewidth(0.4)
            fft_lines[c].set_color(color=colour)
            fft_lines[c].set_linewidth(0.4)

        # set rainbow colours for frequency bands
        for n in range(0, 5):
            a = n/4.0
            if brightness < 0.5:
                colour = (0.3+0.7*(1 - a), 0.5+0.5*(1.0 - 2.0*abs(a - 0.5)), 0.3+0.7*a)
            else:
                colour = (0.7 - 0.7 * (1 - a), 0.5 - 0.5 * (1.0 - 2.0 * abs(a - 0.5)), 0.7 - 0.7 * a)
            bnd_bars[n].set(color=colour)
            hst_lines[n].set_color(color=colour)
            hst_lines[n].set_linewidth(0.5)

        while self._streaming and self._showing_eegpsd:
            try:
                with self._eeg_lock:
                    eeg_max = self._eeg_data.max()
                    eeg_min = self._eeg_data.min()
                    for c in range(0, len(eeg_lines)):
                        eeg_lines[c].set_ydata(self._eeg_data[c]/self._eeg_channel_height + float(self._channels - c))
                self._eeg_channel_height = 0.5*(eeg_max - eeg_min)
                with self._fft_lock:
                    self._fft_channel_height = 0.5*(self._fft_data.max() - self._fft_data.min())
                    for c in range(0, len(fft_lines)):
                        fft_lines[c].set_ydata(self._fft_data[c]/self._fft_channel_height + float(self._channels - c))
                with self._hst_lock:
                    hst_height = self._hst_data.max()
                    hst_height = hst_height or 1.0
                    for b in range(0, len(hst_lines)):
                        hst_lines[b].set_ydata(self._hst_data[b] / hst_height + float(b))
                with self._bnd_lock:
                    for b in range(0, len(bnd_bars)):
                        bnd_bars[b].set_height(self._bnd_data[b] / hst_height)
                self._eeg_canvas.draw()
                self._fft_canvas.draw()
                self._bnd_canvas.draw()
                self._hst_canvas.draw()
            except Exception as e:
                print(e)
                time.sleep(0.5)

        # done.
        print("Ending EEG/PSD display.")

    def _read_lsl(self):
        """
        Read data into buffer a, then call a new thread
        :return:
        """

        # Begin
        print("Starting LSL reading.")

        # init data buffers
        self._eeg_stream.filter(2, 70)
        self._eeg_stream.notch_filter(50)
        self._eeg_stream.get_data()  # reset the number of new samples after the filter is applied
        with self._eeg_lock:
            self._eeg_data = np.zeros((self._channels, self._samples))
        with self._fft_lock:
            self._fft_data = np.zeros((self._channels, int(self._samples/2)))
        with self._bnd_lock:
            self._bnd_data = np.zeros(5)
        with self._hst_lock:
            self._hst_data = np.zeros((5, int(self._history_length * 5.0)))  # history length * update rate of the analysis thread

        # start streaming loop
        self._lsl_label.setEnabled(True)
        while self._streaming:
            with self._eeg_lock:
                self._eeg_data, ts = self._eeg_stream.get_data()
            with self._gui_lock:
                self._lsl_info = "LSL Time {:0.1f}".format(ts[-1])
            #time.sleep(0.1)

        # done.
        print("Ending LSL reading.")

    def _simulate_eeg(self):
        """
        Simulate EEG and write into buffer
        :return:
        """

        # Begin
        print("Starting simulation.")

        # init data buffers
        with self._eeg_lock:
            self._eeg_data = np.zeros((self._channels, self._samples))
        with self._fft_lock:
            self._fft_data = np.zeros((self._channels, int(self._samples / 2)))
        with self._bnd_lock:
            self._bnd_data = np.zeros(5)
        with self._hst_lock:
            self._hst_data = np.zeros(
                (5, int(self._history_length * 5.0)))  # history length * update rate of the analysis thread

        # init noise sources (sine waves of known frequencies)
        samples = 2 * int(self._sampling_rate)
        freqs = np.round((np.cumsum(np.cumsum(np.ones(40))))*39/820+1)     # going from 1 Hz to 40 Hz, staying longer at the low ones
        waves = np.sin(np.cumsum(2.0*np.pi*np.ones((40, samples))/(self._sampling_rate), axis=1).T * freqs).T

        # start streaming loop
        self._lsl_label.setEnabled(True)
        f_window = np.array([0,0,0,0,0])
        t_window = np.arange(int(samples/10))
        cycle = 0
        metacycle = 0
        while self._streaming:
            with self._eeg_lock:
                self._eeg_data = np.roll(self._eeg_data, -int(samples/10))
                self._eeg_data[:, -int(samples/10):] = (waves[f_window, :])[:, t_window]
            with self._gui_lock:
                self._lsl_info = "SML Time {:0.1f} s".format(1.0)
            t_window = np.mod(t_window + int(samples/10), samples)
            cycle += 1
            if cycle == 10:
                cycle = 0
                metacycle += 1
                f_window = np.mod(f_window+1, 40)
                t_window = np.arange(int(samples/10))
            if metacycle == 40:
                metacycle = 0
                f_window = np.mod(f_window + np.array([0,1,2,3,4]), 40)
            time.sleep(0.1)

        # done.
        print("Ending LSL reading.")

    def _analyse_psd(self):
        """
        Read data into buffer a, then call a new thread
        :return:
        """

        # start
        print("Starting analysis.")
        while not len(self._eeg_data):
            time.sleep(0.5)

        # activate labels
        self._eeg_label.setEnabled(True)
        self._bnd_label.setEnabled(True)

        # initialise data objects
        self._fft_freqs = np.fft.rfftfreq(self._samples, d=1.0 / self._sampling_rate)
        bin_freqs = np.array([3.5, 7.5, 12.5, 30.5, 50.0, 70.0])   # delta, theta, alpha, beta, gamma, total
        bins = [abs(self._fft_freqs - f).argmin() for f in bin_freqs]
        widths = np.insert(bin_freqs[1:] - bin_freqs[:-1], 0, bin_freqs[0])
        self._fft_resolution = self._fft_freqs[1]
        smooth = self._bnd_smoothing
        is_relative = True
        is_relative_total = True
        normalisation = [1.0, 2.0, 4.0, 8.0, 16.0]
        is_normalised = True

        # start streaming loop
        while self._streaming:
            try:
                with self._fft_lock:
                    with self._eeg_lock:
                        eeg_min = self._eeg_data.min()
                        eeg_max = self._eeg_data.max()
                        self._fft_data = np.fft.rfft(self._eeg_data, axis=1)
                    self._fft_data = np.abs(self._fft_data)**2
                    fft_all_channels = self._fft_data.sum(axis=0)[1:] / self._channels     # sum of fft over all channels, excluding DC
                    c = self._fft_resolution     # normalise each band by its width, as if it were 1.0 wide
                    if is_relative:
                        bnd_data = np.array([a[0].sum() * c / a[1] for a in
                                             zip(np.split(fft_all_channels, bins)[:5], widths)])
                    else:
                        bnd_data = np.array([a.sum() * c for a in np.split(fft_all_channels, bins)[:5]])
                    if is_normalised:
                        bnd_data = bnd_data * normalisation
                    if is_relative_total:
                        bnd_data = bnd_data / (bnd_data.sum() or 1.0)   # relative power
                    with self._bnd_lock:
                        self._bnd_data = smooth*self._bnd_data + (1.0-smooth)*bnd_data
                        with self._hst_lock:
                            self._hst_data[:, :-1] = self._hst_data[:, 1:]
                            self._hst_data[:, -1] = self._bnd_data
                    if is_relative_total:
                        self._bnd_info = "Freq δ {:0.1f} | θ {:0.1f} | α {:0.1f} | β {:0.1f} | γ {:0.1f}".format(*self._bnd_data)
                    else:
                        self._bnd_info = "Freq δ {:0.1f} | θ {:0.1f} | α {:0.1f} | β {:0.1f} | γ {:0.1f}".format(
                            self._bnd_data[0] * 1e6, self._bnd_data[1] * 1e6, self._bnd_data[2] * 1e6,
                            self._bnd_data[3] * 1e6,
                            self._bnd_data[4] * 1e6)
                    self._eeg_info = "{:.1f} µV - {:.1f} µV".format(eeg_min * 1e6, eeg_max * 1e6)
            except Exception as e:
                print(e)
                time.sleep(0.5)
            time.sleep(0.1)

        # done.
        print("Ending analysis.")

    def _find_sources(self):
        while self._resolving:
            time.sleep(1.0)
            if not self._streaming:
                streams = resolve_streams(timeout=1)
                with self._gui_lock:
                    self._available_streams = ["LSL {} | {} | {} Channels | {} Hz"
                                               "".format(s.name, s.source_id, s.n_channels, s.sfreq)
                                               for s in streams] + ["SML Simulated data / selftest"]

    def _update_gui(self):
        with self._gui_lock:
            if self._streaming:
                self._lsl_label.setText(self._lsl_info)
                self._bnd_label.setText(self._bnd_info)
                self._eeg_label.setText(self._eeg_info)
            else:
                entries = []
                for n in range(self._combobox_streamname.count()):
                    entries += [self._combobox_streamname.itemText(n)]
                for e in entries:
                    if e not in self._available_streams:
                        self._combobox_streamname.removeItem(self._combobox_streamname.findText(e))
                for s in self._available_streams:
                    if s not in entries:
                        self._combobox_streamname.addItem(s)

    def get_data(self):
        return self._bnd_data


class SamadhiWindow(QtWidgets.QMainWindow, Ui_MainWindow):
    """
    Implements the main part of the GUI.
    """

    _minds = []

    def __init__(self, filename=None):
        """
        Initialise the application, connect all button signals to application functions, load theme and last file
        """

        print("starting Samadhi...")

        # initialise main window
        super().__init__()
        self.setupUi(self)

        # connect all button signals
        print("setting up GUI...")
        #
        # ...

        # init application settings
        #print("loading system settings...")
        #self.settings = QtCore.QSettings('FreeSoftware', 'Samadhi')

        # add one mind
        self.add_mind(self.comboBoxStreamName01, self.lineEditName01, self.checkBoxConnect01,
                      self.checkBoxDspEegPsd, self.checkBoxDspDancingDots,
                      self.labelLslStatus, self.labelEegStatus, self.labelFrequencyBands)

    def __del__(self):
        print("Closing main window... ")
        for m in self._minds:
            m.__del__()
        print("... main windows closed.")

    def add_mind(self, combobox_streamname, lineedit_name, checkbox_connect,
                       checkbox_display_eegpsd, checkbox_display_ddots,
                       lsl_info, eeg_info, bnd_info):
        self._minds.append(Mind(combobox_streamname, lineedit_name, checkbox_connect,
                                checkbox_display_eegpsd, checkbox_display_ddots,
                                lsl_info, eeg_info, bnd_info, self.tabWidget))


class Samadhi:

    def __init__(self, filename=None):

        format = QSurfaceFormat()
        format.setProfile(QSurfaceFormat.OpenGLContextProfile.CoreProfile)
        format.setVersion(3, 3)
        QSurfaceFormat.setDefaultFormat(format)
        app = QtWidgets.QApplication(sys.argv)

        # start main window
        main_window = SamadhiWindow()

        # run
        main_window.show()
        sys.exit(app.exec())



# Form implementation generated from reading ui file 'mainwindow.ui'
#
# Created by: PyQt6 UI code generator 6.8.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1257, 900)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon.fromTheme("emblem-shared")
        MainWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.formLayout = QtWidgets.QFormLayout(self.centralwidget)
        self.formLayout.setObjectName("formLayout")
        self.toolBox = QtWidgets.QToolBox(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolBox.sizePolicy().hasHeightForWidth())
        self.toolBox.setSizePolicy(sizePolicy)
        self.toolBox.setMinimumSize(QtCore.QSize(400, 0))
        self.toolBox.setObjectName("toolBox")
        self.pageSource = QtWidgets.QWidget()
        self.pageSource.setGeometry(QtCore.QRect(0, 0, 400, 824))
        self.pageSource.setObjectName("pageSource")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.pageSource)
        self.gridLayout_3.setObjectName("gridLayout_3")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.gridLayout_3.addItem(spacerItem, 3, 0, 1, 1)
        self.groupBox = QtWidgets.QGroupBox(parent=self.pageSource)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.labelEegStatus = QtWidgets.QLabel(parent=self.groupBox)
        self.labelEegStatus.setEnabled(False)
        self.labelEegStatus.setObjectName("labelEegStatus")
        self.gridLayout_4.addWidget(self.labelEegStatus, 5, 1, 1, 1)
        self.comboBoxStreamName01 = QtWidgets.QComboBox(parent=self.groupBox)
        self.comboBoxStreamName01.setObjectName("comboBoxStreamName01")
        self.gridLayout_4.addWidget(self.comboBoxStreamName01, 3, 1, 1, 1)
        self.labelLslStatus = QtWidgets.QLabel(parent=self.groupBox)
        self.labelLslStatus.setEnabled(False)
        self.labelLslStatus.setObjectName("labelLslStatus")
        self.gridLayout_4.addWidget(self.labelLslStatus, 4, 1, 1, 1)
        self.checkBoxConnect01 = QtWidgets.QCheckBox(parent=self.groupBox)
        self.checkBoxConnect01.setObjectName("checkBoxConnect01")
        self.gridLayout_4.addWidget(self.checkBoxConnect01, 3, 0, 1, 1)
        self.lineEditName01 = QtWidgets.QLineEdit(parent=self.groupBox)
        self.lineEditName01.setText("")
        self.lineEditName01.setObjectName("lineEditName01")
        self.gridLayout_4.addWidget(self.lineEditName01, 0, 1, 1, 1)
        self.labelFrequencyBands = QtWidgets.QLabel(parent=self.groupBox)
        self.labelFrequencyBands.setEnabled(False)
        self.labelFrequencyBands.setObjectName("labelFrequencyBands")
        self.gridLayout_4.addWidget(self.labelFrequencyBands, 6, 1, 1, 1)
        self.gridLayout_3.addWidget(self.groupBox, 0, 0, 1, 1)
        self.groupBoxVisualisation = QtWidgets.QGroupBox(parent=self.pageSource)
        self.groupBoxVisualisation.setObjectName("groupBoxVisualisation")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.groupBoxVisualisation)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.checkBoxDspEegPsd = QtWidgets.QCheckBox(parent=self.groupBoxVisualisation)
        self.checkBoxDspEegPsd.setEnabled(False)
        self.checkBoxDspEegPsd.setChecked(False)
        self.checkBoxDspEegPsd.setObjectName("checkBoxDspEegPsd")
        self.gridLayout_6.addWidget(self.checkBoxDspEegPsd, 0, 0, 1, 1)
        self.checkBoxDspDancingDots = QtWidgets.QCheckBox(parent=self.groupBoxVisualisation)
        self.checkBoxDspDancingDots.setEnabled(False)
        self.checkBoxDspDancingDots.setObjectName("checkBoxDspDancingDots")
        self.gridLayout_6.addWidget(self.checkBoxDspDancingDots, 3, 0, 1, 1)
        self.gridLayout_3.addWidget(self.groupBoxVisualisation, 2, 0, 1, 1)
        self.toolBox.addItem(self.pageSource, "")
        self.pageSettings = QtWidgets.QWidget()
        self.pageSettings.setGeometry(QtCore.QRect(0, 0, 400, 824))
        self.pageSettings.setObjectName("pageSettings")
        self.gridLayout_31 = QtWidgets.QGridLayout(self.pageSettings)
        self.gridLayout_31.setObjectName("gridLayout_31")
        self.label = QtWidgets.QLabel(parent=self.pageSettings)
        self.label.setObjectName("label")
        self.gridLayout_31.addWidget(self.label, 0, 0, 1, 1)
        self.toolBox.addItem(self.pageSettings, "")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.toolBox)
        self.tabWidget = QtWidgets.QTabWidget(parent=self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self.tabWidget)
        MainWindow.setCentralWidget(self.centralwidget)
        self.actionQuit = QtGui.QAction(parent=MainWindow)
        self.actionQuit.setObjectName("actionQuit")

        self.retranslateUi(MainWindow)
        self.toolBox.setCurrentIndex(0)
        self.tabWidget.setCurrentIndex(-1)
        self.actionQuit.triggered.connect(MainWindow.close) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Samadhi"))
        self.groupBox.setTitle(_translate("MainWindow", "Data Source"))
        self.labelEegStatus.setText(_translate("MainWindow", "EEG Status"))
        self.labelLslStatus.setText(_translate("MainWindow", "LSL Status"))
        self.checkBoxConnect01.setText(_translate("MainWindow", "Click to connect"))
        self.lineEditName01.setPlaceholderText(_translate("MainWindow", "Name of Person"))
        self.labelFrequencyBands.setText(_translate("MainWindow", "Frequency Bands"))
        self.groupBoxVisualisation.setTitle(_translate("MainWindow", "Visualisation"))
        self.checkBoxDspEegPsd.setText(_translate("MainWindow", "EEG and Spectrum"))
        self.checkBoxDspDancingDots.setText(_translate("MainWindow", "Dancing Dots"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.pageSource), _translate("MainWindow", "Data Source"))
        self.label.setText(_translate("MainWindow", "<html><head/><body><p>Samadhi EEG<br/><br/>Version 2025.2<br/><br/>Software to explore, test, and make available<br/>different visualisations for EEG.<br/><br/>License: GPLv4<br/><br/>Web: <a href=\"https://samadhi-eeg.info\"><span style=\" text-decoration: underline; color:#1d99f3;\">https://samadhi-eeg.info<br/></span></a>Source code, bugs, issues: <a href=\"https://github.com/jkanev/samadhi\"><span style=\" text-decoration: underline; color:#1d99f3;\">github.com/jkanev/samadhi</span></a></p></body></html>"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.pageSettings), _translate("MainWindow", "About"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))



# -*- coding:utf-8 -*-

import sys

# main starting file for use with or with pyinstaller for creating executable binaries.
# For normal running please execute the software as a module (python3 -m samadhi);
import faulthandler
from samadhi import Samadhi

faulthandler.enable()
Samadhi()
